
   /**-------------------------------------------------------------------**
    **                               CLooG                               **
    **-------------------------------------------------------------------**
    **                             domain.c                              **
    **-------------------------------------------------------------------**
    **                   First version: october 28th 2001                **
    **-------------------------------------------------------------------**/


/******************************************************************************
 *               CLooG : the Chunky Loop Generator (experimental)             *
 ******************************************************************************
 *                                                                            *
 * Copyright (C) 2001 Cedric Bastoul                                          *
 *                                                                            *
 * This is free software; you can redistribute it and/or modify it under the  *
 * terms of the GNU General Public License as published by the Free Software  *
 * Foundation; either version 2 of the License, or (at your option) any later *
 * version.                                                                   *
 *                                                                            *
 * This software is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with software; if not, write to the Free Software Foundation, Inc.,        *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA                     *
 *                                                                            *
 * CLooG, the Chunky Loop Generator                                           *
 * Written by Cedric Bastoul, Cedric.Bastoul@inria.fr                         *
 *                                                                            *
 ******************************************************************************/
/* CAUTION: the english used for comments is probably the worst you ever read,
 *          please feel free to correct and improve it !
 */


# include <stdlib.h>
# include <stdio.h>
# include <ctype.h>
# include "../include/cloog/cloog.h"


/******************************************************************************
 *                             Memory leaks hunting                           *
 ******************************************************************************/


/* These functions and global variables are devoted to memory leaks hunting: we
 * want to know at each moment how many Polyhedron structures had been allocated
 * (cloog_domain_allocated) and how many had been freed (cloog_domain_freed).
 * Each time a Polyhedron structure is allocated, a call to the function
 * cloog_domain_leak_up() must be carried out, and respectively
 * cloog_domain_leak_down() when a Polyhedron structure is freed. The special
 * variable cloog_domain_max gives the maximal number of Polyhedron structures
 * simultaneously alive (i.e. allocated and non-freed) in memory.
 * July 3rd->11th 2003: first version (memory leaks hunt and correction).
 */


int cloog_domain_allocated = 0 ;
int cloog_domain_freed = 0 ;
int cloog_domain_max = 0 ;


void cloog_domain_leak_up()
{ cloog_domain_allocated ++ ;
  if ((cloog_domain_allocated-cloog_domain_freed) > cloog_domain_max)
  cloog_domain_max = cloog_domain_allocated - cloog_domain_freed ;
}


void cloog_domain_leak_down()
{ cloog_domain_freed ++ ;
}


/******************************************************************************
 *                              PolyLib interface                             *
 ******************************************************************************/


/* CLooG makes an intensive use of polyhedral operations and the PolyLib do
 * the job. Here are the interfaces to all the PolyLib calls (CLooG uses 18
 * PolyLib functions), with or without some adaptations. If another polyhedral
 * library can be used, only these functions have to be changed. In most of
 * them, there is a call to cloog_domain_compact in order to not use the
 * PolyLib MAX_RAYS number of rays (i.e. number of rows in the rays array) but
 * exactly the number of rays needed in order to save memory. This function
 * may have to be changed or removed too with another polyhedral toolset.
 */


/* cloog_domain_matrix_print function:
 * This function prints the content of a Matrix structure (matrix) into a
 * file (foo, possibly stdout).
 */
void cloog_domain_matrix_print(FILE * foo, Matrix * matrix)
{ Matrix_Print(foo,P_VALUE_FMT,matrix) ;
}


/* cloog_domain_matrix_free function:
 * This function frees the allocated memory for a Matrix structure (matrix).
 */
void cloog_domain_matrix_free(Matrix * matrix)
{ Matrix_Free(matrix) ;
}


/* cloog_domain_matrix_alloc function:
 * This function allocates the memory space for a Matrix structure having
 * nb_rows rows and nb_columns columns, it set its elements to 0.
 */
Matrix * cloog_domain_matrix_alloc(unsigned nb_rows, unsigned nb_columns)
{ return Matrix_Alloc(nb_rows,nb_columns) ;
}


/* cloog_domain_matrix2domain function:
 * Given a matrix of constraints (matrix), this function constructs and returns
 * the corresponding polyhedron (i.e. the double representation: constraint
 * matrix and the set of rays).
 */ 
Polyhedron * cloog_domain_matrix2domain(Matrix * matrix)
{ cloog_domain_leak_up() ;
  return (Constraints2Polyhedron(matrix,MAX_RAYS)) ;
  /*return cloog_domain_compact(Constraints2Polyhedron(matrix,MAX_RAYS)) ;*/
}


/* cloog_domain_print function:
 * This function prints the content of a Polyhedron structure (polyhedron) into
 * a file (foo, possibly stdout).
 */
void cloog_domain_print(FILE * foo, Polyhedron * polyhedron)
{ Polyhedron_Print(foo,P_VALUE_FMT,polyhedron) ;
}


/* cloog_domain_free function:
 * This function frees the allocated memory for a Polyhedron structure
 * (polyhedron), and frees its next polyhedra.
 */
void cloog_domain_free(Polyhedron * polyhedron)
{ if (polyhedron != NULL)
  { cloog_domain_leak_down() ;
    Domain_Free(polyhedron) ;
  }
}


/* cloog_domain_domain2matrix function:
 * Given a polyhedron (polyhedron), this function returns its corresponding
 * matrix of constraints.
 */
Matrix * cloog_domain_domain2matrix(Polyhedron * polyhedron)
{ return Polyhedron2Constraints(polyhedron) ;
}


/* cloog_domain_copy function:
 * This function returns a copy of a polyhedral domain (polyhedron).
 */ 
Polyhedron * cloog_domain_copy(Polyhedron * polyhedron)
{ cloog_domain_leak_up() ;
  return Domain_Copy(polyhedron) ;
}


/* cloog_domain_image function:
 * This function transforms a polyhedral domain (polyhedron) into another
 * domain according to a given affine mapping function (mapping). 
 */
Polyhedron * cloog_domain_image(Polyhedron * polyhedron, Matrix * mapping)
{ cloog_domain_leak_up() ;
  return (DomainImage(polyhedron,mapping,MAX_RAYS)) ;
  /*return cloog_domain_compact(DomainImage(polyhedron,mapping,MAX_RAYS)) ;*/
}


/* cloog_domain_preimage function:
 * Given a polyhedral domain (polyhedron) and a mapping function (mapping)
 * this function returns the polyhedral domain which when transformed by
 * mapping function (mapping) gives (polyhedron).
 */
Polyhedron * cloog_domain_preimage(Polyhedron * polyhedron, Matrix * mapping)
{ cloog_domain_leak_up() ;
  return (DomainPreimage(polyhedron,mapping,MAX_RAYS)) ;
  /*return cloog_domain_compact(DomainPreimage(polyhedron,mapping,MAX_RAYS)) ;*/
}


/* cloog_domain_convex function:
 * Given a polyhedral domain (polyhedron), this function concatenates the lists
 * of rays and lines of the two (or more) polyhedra in the domain into one
 * combined list, and  find the set of constraints which tightly bound all of
 * those objects. It returns the corresponding polyhedron.
 */ 
Polyhedron * cloog_domain_convex(Polyhedron * polyhedron)
{ cloog_domain_leak_up() ;
  return (DomainConvex(polyhedron,MAX_RAYS));
  /*return cloog_domain_compact(DomainConvex(polyhedron,MAX_RAYS));*/
}


/* cloog_domain_simplify function:
 * Given two polyhedral domains (pol1) and (pol2), this function finds the
 * largest domain set (or the smallest list of non-redundant constraints),
 * that when intersected with polyhedral domain (pol2) equals
 * (Pol1)intersect(Pol2). The output is a polyhedral domain with the "redundant"
 * constraints removed.
 * NB: this function do not work as expected with unions of polyhedra...
 */ 
Polyhedron * cloog_domain_simplify(Polyhedron * pol1, Polyhedron * pol2)
{ cloog_domain_leak_up() ;
  return (DomainSimplify(pol1,pol2,MAX_RAYS)) ;
  /*return cloog_domain_compact(DomainSimplify(pol1,pol2,MAX_RAYS)) ;*/
}


/* cloog_domain_union function:
 * This function returns the union of two polyhedral domains (pol1) U (pol2).
 * It returns the resulting domain.
 */
Polyhedron * cloog_domain_union(Polyhedron * pol1, Polyhedron * pol2)
{ cloog_domain_leak_up() ;
  return (DomainUnion(pol1,pol2,MAX_RAYS)) ;
  /*return cloog_domain_compact(DomainUnion(pol1,pol2,MAX_RAYS)) ;*/
}


/* cloog_domain_intersection function:
 * This function returns a pointer to the Polyhedron intersection between two
 * polyhedron given as input: (pol1) inter (pol2).
 */ 
Polyhedron * cloog_domain_intersection(Polyhedron * pol1, Polyhedron * pol2)
{ cloog_domain_leak_up() ;
  return (DomainIntersection(pol1,pol2,MAX_RAYS)) ;
  /*return cloog_domain_compact(DomainIntersection(pol1,pol2,MAX_RAYS)) ;*/
}


/* cloog_domain_difference function:
 * This function returns a pointer to the Polyhedron difference between two
 * polyhedron given as input: domain \ minus.
 * November 8th 2001: first version.
 */ 
Polyhedron * cloog_domain_difference(Polyhedron * domain, Polyhedron * minus)
{ if (cloog_domain_empty(minus))
  return(cloog_domain_copy(domain)) ;
  else
  { cloog_domain_leak_up() ;
    return (DomainDifference(domain,minus,MAX_RAYS)) ;
    /*return(cloog_domain_compact(DomainDifference(domain,minus,MAX_RAYS))) ;*/
  }
}


/* cloog_domain_includes function:
 * This function returns 1 if the polyhedral domain 'container' includes
 * the polyhedral domain 'contents', 0 otherwise.
 * September 14th 2002: first version. 
 */
int cloog_domain_includes(Polyhedron * container, Polyhedron * contents)
{ int is_in ;
  Polyhedron * p1, * p2 ;
  
  for (p1=container; p1; p1=p1->next)
  { is_in = 0 ;
    
    for (p2=contents; p2; p2=p2->next)
    if (PolyhedronIncludes(p1,p2))
    { is_in = 1 ;
      break ;
    }
    
    if (!is_in)
    return 0 ;
  }
  
  return 1 ;
}


/* cloog_domain_addconstraints function :
 * This function adds source's constraints to target: for each element of
 * 'target' (i.e. element of the union of polyhedra) it add the constraints of
 * the corresponding element in 'source'.
 * August 10th 2002: first version.
 * Nota bene for future : it is possible that source and target don't have the
 * same number of elements (try iftest2 without non-shared constraint
 * elimination in cloog_loop_separate !). This function is yet another part
 * of the DomainSimplify patching problem...
 */
Polyhedron * cloog_domain_addconstraints(source, target)
Polyhedron * source, * target ;
{ unsigned nb_constraint ;
  Value * constraints ;
  Polyhedron * new, * next, * last ;

  cloog_domain_leak_up() ;
  constraints = source->p_Init ;
  nb_constraint = source->NbConstraints ;
  source = source->next ;
  new = AddConstraints(constraints,nb_constraint,target,MAX_RAYS) ;
  /*new = cloog_domain_compact(new) ;*/
  last = new ;
  next = target->next ;

  while (next != NULL)
  { cloog_domain_leak_up() ; /* TO AVOID !!! (for iftests and loop_scatter()) */
    /* BUG !!! This is actually a bug. I don't know yet how to cleanly avoid
     * the situation where source and target do not have the same number of
     * elements. So this 'if' is an awful trick, waiting for better.
     */
    if (source != NULL)
    { constraints = source->p_Init ;
      nb_constraint = source->NbConstraints ;
      source = source->next ;
    }
    last->next = AddConstraints(constraints,nb_constraint,next,MAX_RAYS) ;
    /*last->next = cloog_domain_compact(last->next) ;*/
    last = last->next ;
    next = next->next ;
  }

  return(new) ;
}


/* cloog_domain_sort function:
 * This function topologically sorts (nb_pols) polyhedra. Here (pols) is a an
 * array of pointers to polyhedra, (nb_pols) is the number of polyhedra,
 * (level) is the level to consider for partial ordering (nb_par) is the
 * parameter space dimension, (permut) if not NULL, is an array of (nb_pols)
 * integers that contains a permutation specification after call in order to
 * apply the topological sorting. 
 */
void cloog_domain_sort(pols, nb_pols, level, nb_par, permut)
Polyhedron ** pols ;
unsigned nb_pols, level, nb_par ;
int * permut ;
{ int * time ;
  
  /* time is an array of (nb_pols) integers to store logical time values. We
   * do not use it, but it is compulsory for PolyhedronTSort.
   */
  time = (int *)malloc(nb_pols*sizeof(int)) ;

  /* PolyhedronTSort will fill up permut (and time). */
  PolyhedronTSort(pols,nb_pols,level,nb_par,time,permut,MAX_RAYS) ;
    
  free(time) ;
}


/******************************************************************************
 *                          Structure display function                        *
 ******************************************************************************/


/* cloog_domain_list_print function:
 * This function prints the content of a CloogDomainList structure into a
 * file (foo, possibly stdout).
 * November 6th 2001: first version.
 */
void cloog_domain_list_print(FILE * foo, CloogDomainList * list)
{ while (list != NULL)
  { cloog_domain_print(foo,list->domain) ;
    list = list->next ;
  }
}


/******************************************************************************
 *                         Memory deallocation function                       *
 ******************************************************************************/


/* cloog_domain_list_free function:
 * This function frees the allocated memory for a CloogDomainList structure.
 * November 6th 2001: first version.
 */
void cloog_domain_list_free(CloogDomainList * list)
{ CloogDomainList * temp ;
  
  while (list != NULL)
  { temp = list->next ;
    cloog_domain_free(list->domain) ;
    free(list) ;
    list = temp ;
  }
}


/******************************************************************************
 *                               Reading function                             *
 ******************************************************************************/


/* cloog_domain_read function:
 * Adaptation from the PolyLib. This function reads a matrix into a file (foo,
 * posibly stdin) and returns a pointer to a polyhedron containing the read
 * information. 
 * October 18th 2001: first version.
 */
Polyhedron * cloog_domain_read(FILE * foo)
{ unsigned NbRows, NbColumns ;
  int i, j, n ;
  char *c, s[MAX_STRING], str[MAX_STRING] ;
  Matrix * matrix ;
  Polyhedron * domain ;
  Value * p ;
  
  while (fgets(s,MAX_STRING,foo) == 0) ;
  while ((*s=='#' || *s=='\n') || (sscanf(s," %u %u",&NbRows,&NbColumns)<2))
  fgets(s,MAX_STRING,foo) ;
  
  matrix = cloog_domain_matrix_alloc(NbRows,NbColumns) ;

  p = matrix->p_Init ;
  for (i=0;i<matrix->NbRows;i++) 
  { do 
    { c = fgets(s,MAX_STRING,foo) ;
      while ((c != NULL) && isspace(*c) && (*c != '\n'))
      c++ ;
    }
    while (c != NULL && (*c == '#' || *c == '\n'));
    
    if (c == NULL) 
    { fprintf(stderr, "[CLooG]ERROR: not enough rows.\n") ;
      exit(1) ;
    }
    for (j=0;j<matrix->NbColumns;j++) 
    { if (c == NULL || *c == '#' || *c == '\n')
      { fprintf(stderr, "[CLooG]ERROR: not enough columns.\n") ;
        exit(1) ;
      }
      /* NdCed : Dans le n ca met strlen(str). */
      if (sscanf(c,"%s%n",str,&n) == 0) 
      { fprintf(stderr, "[CLooG]ERROR: not enough rows.\n") ;
        exit(1) ;
      }
      sscanf(str,VALUE_FMT,p++) ;
      c += n ;
    }
  }
  domain = cloog_domain_matrix2domain(matrix) ;
  cloog_domain_matrix_free(matrix) ;
  
  return(domain) ;
}


/* cloog_domain_union_read function:
 * This function reads a union of polyhedra into a file (foo, posibly stdin) and
 * returns a pointer to a Polyhedron containing the read information. 
 * September 9th 2002: first version.
 */
Polyhedron * cloog_domain_union_read(FILE * foo)
{ int i, nb_components ;
  char s[MAX_STRING] ;
  Polyhedron * domain, * temp, * old ;

  /* domain reading: nb_components (constraint matrices). */
  while (fgets(s,MAX_STRING,foo) == 0) ;
  while ((*s=='#' || *s=='\n') || (sscanf(s," %d",&nb_components)<1))
  fgets(s,MAX_STRING,foo) ;
  
  if (nb_components > 0)
  { /* 1. first part of the polyhedra union, */
    domain = cloog_domain_read(foo) ;
    /* 2. and the nexts. */
    for (i=1;i<nb_components;i++)
    { /* The entire union will be freed by a single call to domain_free, so we
       * want to count only one allocation too.
       */
      cloog_domain_allocated -- ;
      
      temp = cloog_domain_read(foo) ;
      old = domain ;
      domain = cloog_domain_union(temp,old) ;
      cloog_domain_free(temp) ;
      cloog_domain_free(old) ;
    }
    
    return domain ;
  }
  else
  return NULL ;
}


/* cloog_domain_list_read function:
 * This function reads a list of polyhedra into a file (foo, posibly stdin) and
 * returns a pointer to a CloogDomainList containing the read information. 
 * November 6th 2001: first version.
 */
CloogDomainList * cloog_domain_list_read(FILE * foo)
{ int i, nb_pols ;
  char s[MAX_STRING] ;
  CloogDomainList * list, * now, * next ;
  
  
  /* We read first the number of polyhedra in the list. */
  while (fgets(s,MAX_STRING,foo) == 0) ;
  while ((*s=='#' || *s=='\n') || (sscanf(s," %d",&nb_pols)<1))
  fgets(s,MAX_STRING,foo) ;
  
  /* Then we read the polyhedra. */
  list = NULL ;
  if (nb_pols > 0)
  { list = (CloogDomainList *)malloc(sizeof(CloogDomainList)) ;
    list->domain = cloog_domain_read(foo) ;
    list->next = NULL ;
    now = list ;
    for (i=1;i<nb_pols;i++)
    { next = (CloogDomainList *)malloc(sizeof(CloogDomainList)) ;
      next->domain = cloog_domain_read(foo) ;
      next->next = NULL ;
      now->next = next ;
      now = now->next ;
    }
  }
  return(list) ;
}


/******************************************************************************
 *                            Processing functions                            *
 ******************************************************************************/


/* cloog_domain_compact function:
 * This function returns the copy of the polyhedron given as input, but
 * without any unused space. Actually a lot of memory is lost because of the
 * NbMaxRays (i.e. MAX_RAYS in CLooG) number of rays allocated by default by
 * the PolyLib for each polyhedron resulting from a polyhedral operation. This
 * function frees the Polyhedron structure given as input and returns its
 * compacted copy.
 * July 14th 2003: first version.
 */
Polyhedron * cloog_domain_compact(Polyhedron * big)
{ Polyhedron * compact ;
  compact = cloog_domain_copy(big) ;
  cloog_domain_free(big) ;
  return compact ;
}


/* cloog_domain_empty function:
 * This function returns 1 if the polyhedron given as input is empty, 0
 * otherwise.
 * October 28th 2001: first version.
 */ 
int cloog_domain_empty(Polyhedron * polyhedron)
{ if (polyhedron == NULL)
  return 1 ;

  if (polyhedron->next)
  return(0) ;
  return((polyhedron->Dimension < polyhedron->NbEq) ? 1 : 0) ;
}


/* cloog_domain_universe function:
 * This function returns 1 if the polyhedron given as input describe the
 * universe of its dimension, 0 otherwise. Nb: the NbBid field of a polyhedron
 * gives the number of bidirectionnal rays.
 * November 19th 2001: first version.
 */
int cloog_domain_universe(Polyhedron * polyhedron)
{ if (polyhedron->next)
  return(0) ;
  return((polyhedron->Dimension == polyhedron->NbBid) ? 1 : 0) ;
}


/* cloog_domain_project function:
 * From Quillere's LoopGen 0.4. This function returns the projection of
 * (domain) on the (level) first dimensions (i.e. outer loops). It returns a
 * pointer to the projected Polyhedron.
 * - nb_par is the number of parameters.
 * (October 27th 2001)
 */ 
Polyhedron * cloog_domain_project(Polyhedron * domain, int level, int nb_par)
{ int row, column, nb_rows, nb_columns, difference ;
  Polyhedron * projected_domain ;
  Matrix * matrix ;

  nb_rows = level + nb_par + 1 ;
  nb_columns = domain->Dimension + 1 ;
  difference = nb_columns - nb_rows ;
  
  if (difference == 0)
  return(cloog_domain_copy(domain)) ;
  
  matrix = cloog_domain_matrix_alloc(nb_rows,nb_columns) ;
     
  for (row=0;row<level;row++)
  for (column=0;column<nb_columns; column++)
  matrix->p[row][column] = (row == column ? 1 : 0) ;

  for (;row<nb_rows;row++)
  for (column=0;column<nb_columns;column++)
  matrix->p[row][column] = (row+difference == column ? 1 : 0) ;
  
  projected_domain = cloog_domain_image(domain,matrix) ;
  cloog_domain_matrix_free(matrix) ;

  return(projected_domain) ;
}  


/* cloog_domain_extend function:
 * From Quillere's LoopGen 0.4. This function returns the (domain) given as
 * input with (dim)+(nb_par) dimensions. The new dimensions are added before
 * the (nb_par) parameters. This function does not free (domain), and returns
 * a new polyhedron.
 * - nb_par is the number of parameters.
 * (October 27th 2001)
 */ 
Polyhedron * cloog_domain_extend(Polyhedron * domain, int dim, int nb_par)
{ int row, column, nb_rows, nb_columns, difference ;
  Polyhedron * extended_domain ;
  Matrix * matrix ;
  
  nb_rows = 1 + domain->Dimension ;
  nb_columns = dim + nb_par + 1 ;
  difference = nb_columns - nb_rows ;
  
  if (difference == 0)
  return(cloog_domain_copy(domain)) ;
  
  matrix = cloog_domain_matrix_alloc(nb_rows,nb_columns) ;
    
  for (row=0;row<domain->Dimension-nb_par;row++)
  for (column=0;column<nb_columns;column++)
  matrix->p[row][column] = (row == column ? 1 : 0) ;
  
  for (;row<=domain->Dimension;row++)
  for (column=0;column<nb_columns;column++)
  matrix->p[row][column] = (row+difference == column ? 1 : 0) ;
  
  extended_domain = cloog_domain_preimage(domain,matrix) ;
  cloog_domain_matrix_free(matrix) ;

  return(extended_domain) ;
}


/* cloog_domain_never_integral function:
 * For us, an equality like 3*i -4 = 0 is always false since 4%3 != 0. This
 * function returns a boolean set to 1 if there is this kind of 'never true'
 * constraint inside a polyhedron, 0 otherwise.
 * - domain is the polyhedron to check,
 * - level is the column number of the unknown we want to be integral,
 * - nb_par is the number of parameters in the domain unknowns.
 * November 28th 2001: first version. 
 * June 26th 2003: for iterators, more 'never true' constraints are found
 *                 (compare cholesky2 and vivien with a previous version),
 *                 checking for the parameters created (compare using vivien).
 * June 28th 2003: Previously in loop.c and called
 *                 cloog_loop_simplify_nevertrue, now here !
 */
int cloog_domain_never_integral(Polyhedron * domain, int level, int nb_par)
{ int i, j, level_par, dimension ;

  dimension = domain->Dimension + 2 ;
  
  /* Check for the iterators. */
  for (i=0;i<domain->NbConstraints;i++)  
  { if (!domain->Constraint[i][0] && domain->Constraint[i][dimension-1] &&
        domain->Constraint[i][level])
    { for (j=1;j<dimension-1;j++)  
      { if ((domain->Constraint[i][j] != 0) &&
	    (domain->Constraint[i][j]%domain->Constraint[i][level] != 0))
        break ;
      }
      if (j == dimension-1)
      if (domain->Constraint[i][dimension-1]%domain->Constraint[i][level])
      return(1) ;
    }
  }

  /* Check for the parameters. */
  for (level_par=dimension-1-nb_par;level_par<dimension-1;level_par++)
  for (i=0;i<domain->NbConstraints;i++)  
  { if (!domain->Constraint[i][0] && domain->Constraint[i][dimension-1] &&
        domain->Constraint[i][level_par])
    { for (j=1;j<dimension-1-nb_par;j++)
      if (domain->Constraint[i][j] != 0)
      break ;
      
      if (j == dimension-1-nb_par)
      { for (;j<dimension-1;j++)  
        { if ((domain->Constraint[i][j] != 0) &&
	      (domain->Constraint[i][j]%domain->Constraint[i][level_par] != 0))
          break ;
        }
        if (j == dimension-1)
        if (domain->Constraint[i][dimension-1]%domain->Constraint[i][level_par])
        return(1) ;
      }
    }
  }
  
  return(0) ;
}


int cloog_domain_gcd(int a, int b)
{ if (b == 0)
  return(a) ;
  else
  return(cloog_domain_gcd(b,a%b)) ;
}


/* cloog_domain_stride function:
 * This function finds the stride imposed to unknown with the column number
 * 'strided_level' in order to be integral. For instance, if we have a
 * constraint like -i - 2j + 2k = 0, and we consider k, then k can be integral
 * only if (i + 2j)%2 = 0. Then only if i%2 = 0. Then k imposes a stride 2 to
 * the unknown i. The function returns the imposed stride.
 * - domain is the set of constraint we have to consider,
 * - strided_level is the column number of the unknown for which a stride have
 *   to be found,
 * - looking_level is the column number of the unknown that impose a stride to
 *   the first unknown.
 * - offset is the value of the constant c if the condition is of the shape
 *   (i + c)%s = 0, s being the stride.
 * June 28th 2003: first version.
 * July 14th 2003: can now look for multiple striding constraints and returns
 *                 the GCD of the strides and the common offset.
 */
Value cloog_domain_stride(domain, strided_level, nb_par, offset)
Polyhedron * domain ;
int strided_level, nb_par ;
Value * offset ;
{ int i, j, valid, constant, looking_level, looking_max ;
  Value looking_coeff, strided_coeff, ref_offset=0, potential=0, stride=0 ;

  /* Looking for a non-unit stride. */
  looking_max = domain->Dimension - nb_par + 1 ;
  for (looking_level=strided_level+1;looking_level<looking_max;looking_level++)
  { for (i=0;i<domain->NbConstraints;i++)  
    { strided_coeff = domain->Constraint[i][strided_level] ;
      looking_coeff = domain->Constraint[i][looking_level] ;
    
      /* A potential interesting constraint is an equality such as the
       * coefficient of the potentially non-unit strided iterator and the
       * looking one are not zeros, and such as the looking coefficient do
       * not divide the strided one.
       */
      valid = 1 ;
      if (!domain->Constraint[i][0] && strided_coeff && looking_coeff &&
          (strided_coeff%looking_coeff))
      { /* We verify that all the coefficient except the strided and the
         * looking one give 0 by modulo looking_coef.
         */
        for (j=1;j<strided_level;j++)  
        if (domain->Constraint[i][j]%looking_coeff)
        { valid = 0 ;
          break ;
        }
	    
        if (valid)
        for (j=looking_level+1;j<=domain->Dimension;j++)
        if (domain->Constraint[i][j]%looking_coeff)
        { valid = 0 ;
          break ;
        }

        /* If there is actually a non-unit stride, we take its absolute value.*/
        if (valid)
        { constant = domain->Constraint[i][domain->Dimension+1] ;
	  if (!(*offset=constant%looking_coeff))
	  *offset = 0 ;
	
          potential = (looking_coeff > 0) ? looking_coeff : -1 * looking_coeff ;
        }
      }
    }

    /* If a stride was found, we have to verify that the offset and the
     * stride values are compatible with previous values (offsets must be the
     * same, and we have to consider the GCD of the strides.
     */
    if (potential)
    { if (stride)
      { if (*offset != ref_offset)
        { *offset = 0 ;
	  stride = 1 ;
	  break ;
	}
	else
        stride = cloog_domain_gcd(stride,potential) ;
      }
      else
      { /* We set the reference values. */
        stride = potential ;
        ref_offset = *offset ;
      }
      potential = 0 ;
    }
  }

  /* If no non-unit stride was found, this is because the stride is 1. */
  if (!stride)
  { *offset = 0 ;
    stride = 1 ;
  }
  
  return stride ;
}


/* cloog_domain_integral_lowerbound function:
 * This function returns 1 if the lower bound of an iterator (such as its
 * column rank in the constraint set 'domain' is 'level') is integral,
 * 0 otherwise. If the lower bound is actually integral, the function fills
 * the 'lower' field with the lower bound value.
 * Jun 29th 2003: first version.
 */
int cloog_domain_integral_lowerbound(domain, level, lower)
Polyhedron * domain ;
int level ;
Value * lower ;
{ int i, first_lower=1, lower_constraint=-1 ;
  Value iterator, constant ;
  
  /* We want one and only one lower bound (e.g. no equality, no maximum
   * calculation...).
   */
  for (i=0;i<domain->NbConstraints;i++)  
  if ((!domain->Constraint[i][0]) && (domain->Constraint[i][level]))
  return 0 ;

  for (i=0;i<domain->NbConstraints;i++)  
  if (domain->Constraint[i][level] > 0)
  { if (first_lower)
    { first_lower = 0 ;
      lower_constraint = i ;
    }
    else
    return 0 ;
  }
  if (first_lower)
  return 0 ;
  
  /* We want an integral lower bound: no other non-zero entry except the
   * iterator coefficient and the constant.
   */
  for (i=1;i<level;i++)  
  if (domain->Constraint[lower_constraint][i])
  return 0 ;
  for (i=level+1;i<=domain->Dimension;i++)  
  if (domain->Constraint[lower_constraint][i])
  return 0 ;

  /* If all is passed, then find the lower bound and return 1. */
  iterator = domain->Constraint[lower_constraint][level] ;
  constant = -domain->Constraint[lower_constraint][domain->Dimension+1] ;

  if ((constant%iterator == 0) || (constant < 0))
  *lower = constant/iterator ;
  else
  *lower = constant/iterator + 1;
  
  return 1 ;
}


/* cloog_domain_lowerbound_update function:
 * This function updates the integral lower bound of an iterator (such as its
 * column rank in the constraint set 'domain' is 'level') into  'lower'.
 * Jun 29th 2003: first version.
 */
void cloog_domain_lowerbound_update(domain, level, lower)
Polyhedron * domain ;
int level ;
Value lower ;
{ int i ;
  
  /* There is only one lower bound, the first one is the good one. */
  for (i=0;i<domain->NbConstraints;i++)  
  if (domain->Constraint[i][level] > 0)
  { domain->Constraint[i][level] = 1 ;
    domain->Constraint[i][domain->Dimension+1] = -lower ;
    break ;
  }
}


/* cloog_domain_prism_level function:
 * This function returns the answer to the question: 'how many polyhedron
 * bounds are not constants ?'. 
 * July 30th 2003: first version.
 */
int cloog_domain_prism_level(domain, nb_par)
Polyhedron * domain ;
int nb_par ;
{ int i, j, k, d, non_null=0 ;
  Polyhedron * copy ;

  if (!(domain->Dimension-nb_par))
  return -1 ;

  copy = cloog_domain_copy(domain) ;

  /* Impressive static control kernel, isn't it ?! */
  for (d=1;d<=copy->Dimension-nb_par;d++)
  { for (i=0;i<copy->NbConstraints;i++)
    { if (copy->Constraint[i][d] != 0)
      for (j=d+1;j<=copy->Dimension-nb_par;j++)
      { if (copy->Constraint[i][j] != 0)
        { non_null ++ ;
	  
	  for (k=i;k<copy->NbConstraints;k++)
	  if (copy->Constraint[k][j] != 0)
	  copy->Constraint[k][j] = 0 ;
        }
      }
    }
  }
  
  cloog_domain_free(copy) ;
  return non_null ;
}


/* cloog_domain_point function:
 * This function returns 1 if all the dimensions greater or equal to (level)
 * of (domain) are constants, 0 otherwise. This function do not recognize a
 * constant dimension if it is not explicited with an equality, e.g. if we have
 * d>=0 and d<=0, it will not find the equality d=0.
 * March 27th 2004: first version.
 */
int cloog_domain_point(domain, level, nb_par)
Polyhedron * domain ;
int nb_par ;
{ int i, j, d, equal=1 ;

  /* For each dimension, we have to check if it is constant. */
  for (d=level;d<=domain->Dimension-nb_par;d++)
  { /* We look in each constraint for an equality with a constant for the
     * current dimension.
     */
    for (i=0;i<domain->NbConstraints;i++)
    { equal = 1 ;
      /* We want an equality concerning the current dimension. */
      if ((domain->Constraint[i][0] == 0) && (domain->Constraint[i][d] != 0))
      { /* The equality must not concern other dimensions. */
        for (j=level;j<=d-1;j++)
        { if (domain->Constraint[i][j] != 0)
          { equal = 0 ;
            break ;
          }
        }
        for (j=d+1;j<=domain->Dimension-nb_par;j++)
        { if (domain->Constraint[i][j] != 0)
          { equal = 0 ;
            break ;
          }
        }
      }
      else
      equal = 0 ;
      
      /* We found our equality. */
      if (equal == 1)
      break ;
    }
    /* If there is not such equality for the current dimension, return 0. */
    if (equal == 0)
    return 0 ;
  }
  
  return 1 ;
}


/* cloog_domain_quick_point function:
 * This function returns 1 if the domain given as input is a point, 0 if it
 * is unable to decide. A domain is a point iff there are no dimensions or if
 * each dimension has a fixed value. This function checks that the constraints
 * have the right shape according to the CLooG's users guide: all equalities
 * are given explicitely and such that the iterator dimension part of the
 * constraint matrix is a diagonal matrix, as WRAP-IT does. Trying to proove
 * precisely that the domain is a point is not the aim of this function, it has
 * just to decide as quick as possible for most of the cases.
 * December 15th 2003: first version.
 */
int cloog_domain_quick_point(domain, nb_par)
Polyhedron * domain ;
int nb_par ;
{ int i, j, nb_iterators ;

  nb_iterators = domain->Dimension - nb_par ;
  
  /* If there are no iterators, the domain is obviously a point. */
  if (nb_iterators == 0)
  return 1 ;
  
  /* If there are not enough constraints to define an equality for each
   * iterator, the domain cannot be a point.
   */
  if (domain->NbConstraints < nb_iterators)
  return 0 ;
  
  /* While not guaranteed, if there are too much constraints to define an
   * equality per iterator, there is a lot of chances that the domain is not
   * a point. For this case, we accept one additionnal constraint for the
   * polylib '1>=0'.
   *
  if (domain->NbConstraints > nb_iterators+1)
  return 0 ;*/

  /* Reasoning with the CLooG's scattering functions as described in the
   * user's manual, the constraints should begin with the iterator equalities
   * and this part of the constraint matrix should be a diagonal matrix.
   * We check for this and at the first divergence, we answer we are unable to
   * statuate if the domain is a point or not.
   */
  for (i=0;i<nb_iterators;i++)
  { if (domain->Constraint[i][0] != 0)
    return 0 ;
  
    /* The diagonal element must be != 0. */
    if (domain->Constraint[i][i+1] == 0)
    return 0 ;
    
    /* The elements before the diagonal element must be 0. */
    for (j=1;j<=i;j++)
    if (domain->Constraint[i][j] != 0)
    return 0 ;
    
    /* The elements after the diagonal element must be 0. */
    for (j=i+2;j<=nb_iterators;j++)
    if (domain->Constraint[i][j] != 0)
    return 0 ;
  }
  
  /* All remaining constraints must concern only the parameters. */
  for (i=nb_iterators;i<domain->NbConstraints;i++)
  for (j=1;j<=nb_iterators;j++)
  if (domain->Constraint[i][j] != 0)
  return 0 ;
  

  return 1 ;
}


/* cloog_domain_quick_zero function:
 * This function returns 1 if the domain given as input is a point with
 * coordinate 0, 0 if it is unable to decide. A domain is a point iff there are
 * no dimensions or if each dimension has a fixed value. This function checks
 * that the constraints have the right shape according to the CLooG's users
 * guide: all equalities are given explicitely and such that the iterator
 * dimension part of the constraint matrix is a diagonal matrix, as WRAP-IT
 * does. Trying to proove precisely that the domain is a point is not the aim
 * of this function, it has just to decide as quick as possible for most of the
 * situations.
 * December 15th 2003: first version.
 */
int cloog_domain_quick_zero(domain, nb_par)
Polyhedron * domain ;
int nb_par ;
{ int i, j, nb_iterators ;

  nb_iterators = domain->Dimension - nb_par ;
  
  /* If there are no iterators, the domain is obviously a point. */
  if (nb_iterators == 0)
  return 1 ;
  
  /* If there are not enough constraints to define an equality for each
   * iterator, the domain cannot be a point.
   */
  if (domain->NbConstraints < nb_iterators)
  return 0 ;
  
  /* While not guaranteed, if there are too much constraints to define an
   * equality per iterator, there is a lot of chances that the domain is not
   * a point. For this case, we accept one additionnal constraint for the
   * polylib '1>=0'.
   *
  if (domain->NbConstraints > nb_iterators+1)
  return 0 ;*/

  /* Reasoning with the CLooG's scattering functions as described in the
   * user's manual, the constraints should begin with the iterator equalities
   * and this part of the constraint matrix should be a diagonal matrix.
   * We check for this and at the first divergence, we answer we are unable to
   * statuate if the domain is a point or not.
   */
  for (i=0;i<nb_iterators;i++)
  { if (domain->Constraint[i][0] != 0)
    return 0 ;
  
    /* The diagonal element must be != 0. */
    if (domain->Constraint[i][i+1] == 0)
    return 0 ;
    
    /* The elements before the diagonal element must be 0. */
    for (j=1;j<=i;j++)
    if (domain->Constraint[i][j] != 0)
    return 0 ;
    
    /* The elements after the diagonal element must be 0. */
    for (j=i+2;j<=nb_iterators;j++)
    if (domain->Constraint[i][j] != 0)
    return 0 ;
    
    /* The rest of the constraint must be 0. */
    for (j=nb_iterators+1;j<=domain->Dimension+1;j++)
    if (domain->Constraint[i][j] != 0)
    return 0 ;
  }
  
  /* All remaining constraints must concern only the parameters. */
  for (i=nb_iterators;i<domain->NbConstraints;i++)
  for (j=1;j<=nb_iterators;j++)
  if (domain->Constraint[i][j] != 0)
  return 0 ;
  

  return 1 ;
}


/* cloog_domain_quick_equal function:
 * This function returns 1 if the domains given as input are the same, 0 if it
 * is unable to decide. This function makes an entry-to-entry comparison between
 * the constraint systems, if all the entries are the same, the domains are
 * obviously the same and it returns 1, at the first difference, it returns 0.
 * This is a very fast way to verify this property. It has been shown (with the
 * CLooG benchmarks) that operations on equal domains are 17% of all the
 * polyhedral computations. For 75% of the actually identical domains, this
 * function answer that they are the same and allow to give immediately the
 * trivial solution instead of calling the heavy general functions of PolyLib.
 * August 22th 2003: first version.
 */
int cloog_domain_quick_equal(Polyhedron * p1, Polyhedron * p2)
{ int i, nb_elements ;

  while ((p1 != NULL) && (p2 != NULL))
  { if ((p1->NbConstraints != p2->NbConstraints) ||
        (p1->Dimension != p2->Dimension))
    return 0 ;
  
    nb_elements = p1->NbConstraints * (p1->Dimension + 2) ;
  
    for (i=0;i<nb_elements;i++)
    if (p1->p_Init[i] != p2->p_Init[i])
    return 0 ;
    
    p1 = p1->next ;
    p2 = p2->next ;
  }
  
  if ((p1 != NULL) || (p2 != NULL))
  return 0 ;
  return 1 ;
}


/* cloog_domain_quick_disjoint function:
 * This function returns 1 if the domains given as input are disjoint, 0 if it
 * is unable to decide. This function finds the unknown with fixed values in
 * both domains (on a given constraint, their column entry is not zero and
 * only the constant coefficient can be different from zero) and verify that
 * their values are the same. If not, the domains are obviously disjoint and
 * it returns 1, if there is not such case it returns 0.  This is a very fast
 * way to verify this property. It has been shown (with the CLooG benchmarks)
 * that operations on disjoint domains are 36% of all the polyhedral
 * computations. For 94% of the actually identical domains, this
 * function answer that they are disjoint and allow to give immediately the
 * trivial solution instead of calling the heavy general functions of PolyLib.
 * August 22th 2003: first version.
 */
int cloog_domain_quick_disjoint(Polyhedron * p1, Polyhedron * p2)
{ int i1, j1, i2, j2, scat_dim, scat_val ;

  if ((p1->next != NULL) || (p2->next != NULL))
  return 0 ;
  
  for (i1=0;i1<p1->NbConstraints;i1++)
  { if (p1->Constraint[i1][0] != 0)
    continue ;
    
    scat_dim = 1 ;
    while ((p1->Constraint[i1][scat_dim] == 0) && (scat_dim < p1->Dimension))
    scat_dim ++ ;
    
    if (p1->Constraint[i1][scat_dim] != 1)
    continue ;
    else
    { for (j1=scat_dim+1;j1<=p1->Dimension;j1++)
      if (p1->Constraint[i1][j1] != 0)
      break ;
      
      if (j1 != p1->Dimension+1)
      continue ;
      
      scat_val = p1->Constraint[i1][p1->Dimension+1] ;
            
      for (i2=0;i2<p2->NbConstraints;i2++)
      { for (j2=0;j2<scat_dim;j2++)
        if (p2->Constraint[i2][j2] != 0)
        break ;
       
        if ((j2 != scat_dim) || (p2->Constraint[i2][scat_dim] != 1))
        continue ;
       
        for (j2=scat_dim+1;j2<p2->Dimension;j2++)
        if (p2->Constraint[i2][j2] != 0)
        break ;
       
        if (j2 != p2->Dimension)
        continue ;
       
        if (p2->Constraint[i2][p2->Dimension+1] != scat_val)
        return 1 ;
      }
    }
  }

  return 0 ;
} 
 
 
/* cloog_domain_list_quick_same function:
 * This function returns 1 if two domains in the list are the same, 0 if it
 * is unable to decide.
 * February 9th 2004: first version.
 */
int cloog_domain_list_quick_same(CloogDomainList * list)
{ CloogDomainList * current, * next ;

  current = list ;
  while (current != NULL)
  { next = current->next ;
    while (next != NULL)
    { if (cloog_domain_quick_equal(current->domain,next->domain))
      return 1 ;
      
      next = next->next ;
    }
    current = current->next ;
  }
  
  return 0 ;
}


/* cloog_domain_grow function:
 * This function extend the polyhedron (domain) onto the dimension (level) by a
 * step of 1, if (lower) is 1 then the lower bound of the dimension is the same
 * minus one, if (lower) is 0 then the upper bound of the dimension is the
 * same plus one. This function frees the Polyhedron structure given as input
 * and returns the extended one.
 * March 27th 2004: first version.
 */
Polyhedron * cloog_domain_grow(Polyhedron * domain, int level, int lower)
{ int i ;
  Matrix * matrix ;
  Polyhedron * grow ;
  
  matrix = cloog_domain_domain2matrix(domain) ;
  cloog_domain_free(domain) ;

  for (i=0;i<matrix->NbRows;i++)
  if (matrix->p[i][0] == 1)
  { if (((lower == 1) && (matrix->p[i][level] > 0)) ||
        ((lower == 0) && (matrix->p[i][level] < 0)))
    matrix->p[i][matrix->NbColumns-1] += 1 ;
  }
  
  grow = cloog_domain_matrix2domain(matrix) ;
  cloog_domain_matrix_free(matrix) ;
  return grow ;
}



Polyhedron * cloog_domain_restrict(domain, context, nb_par)
Polyhedron * domain, * context ;
int nb_par ;
{ int new_dimension ;
  Polyhedron * extended_context, * new_domain ;
    
  if (domain->Dimension > context->Dimension)
  { new_dimension = domain->Dimension - nb_par ;
    extended_context = cloog_domain_extend(context,new_dimension,nb_par) ;
    new_domain = cloog_domain_intersection(extended_context,domain) ;
    cloog_domain_free(extended_context) ;
  }
  else
  new_domain = cloog_domain_intersection(context,domain) ;
  
  cloog_domain_free(domain) ;
  return(new_domain) ;
}

