
   /**-------------------------------------------------------------------**
    **                              CLooG                                **
    **-------------------------------------------------------------------**
    **                             loop.c                                **
    **-------------------------------------------------------------------**
    **                  First version: october 26th 2001                 **
    **-------------------------------------------------------------------**/


/******************************************************************************
 *               CLooG : the Chunky Loop Generator (experimental)             *
 ******************************************************************************
 *                                                                            *
 * Copyright (C) 2001 Cedric Bastoul                                          *
 *                                                                            *
 * This is free software; you can redistribute it and/or modify it under the  *
 * terms of the GNU General Public License as published by the Free Software  *
 * Foundation; either version 2 of the License, or (at your option) any later *
 * version.                                                                   *
 *                                                                            *
 * This software is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with software; if not, write to the Free Software Foundation, Inc.,        *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA                     *
 *                                                                            *
 * CLooG, the Chunky Loop Generator                                           *
 * Written by Cedric Bastoul, Cedric.Bastoul@inria.fr                         *
 *                                                                            *
 ******************************************************************************/
/* CAUTION: the english used for comments is probably the worst you ever read,
 *          please feel free to correct and improve it !
 */

# include <stdlib.h>
# include <stdio.h>
# include "../include/cloog/cloog.h"

int qdisjoint = 0, qequal = 0, disjoint = 0, equal = 0, calcul = 0 ;


/******************************************************************************
 *                             Memory leaks hunting                           *
 ******************************************************************************/


/* These functions and global variables are devoted to memory leaks hunting: we
 * want to know at each moment how many CloogLoop structures had been allocated
 * (cloog_loop_allocated) and how many had been freed (cloog_loop_freed).
 * Each time a CloogLoog structure is allocated, a call to the function
 * cloog_loop_leak_up() must be carried out, and respectively
 * cloog_loop_leak_down() when a CloogLoop structure is freed. The special
 * variable cloog_loop_max gives the maximal number of CloogLoop structures
 * simultaneously alive (i.e. allocated and non-freed) in memory.
 * July 3rd->11th 2003: first version (memory leaks hunt and correction).
 */
 
 
int cloog_loop_allocated = 0 ;
int cloog_loop_freed = 0 ;
int cloog_loop_max = 0 ;
extern int cloog_domain_freed ;


void cloog_loop_leak_up()
{ cloog_loop_allocated ++ ;
  if ((cloog_loop_allocated-cloog_loop_freed) > cloog_loop_max)
  cloog_loop_max = cloog_loop_allocated - cloog_loop_freed ;
}


void cloog_loop_leak_down()
{ cloog_loop_freed ++ ;
}


/******************************************************************************
 *                          Structure display function                        *
 ******************************************************************************/


/* cloog_loop_print function:
 * This function prints the content of a CloogLoop structure (start) into a
 * file (foo, possibly stdout).
 */
void cloog_loop_print(FILE * foo, CloogLoop * start)
{ CloogLoop * loop ;

  loop = start ;
  if (loop == NULL)
  fprintf(foo,"Null loop.\n") ;
  
  while (loop != NULL)
  { fprintf(foo,"Domain for the loop (stride %d):\n",loop->stride) ;
    if (loop->domain != NULL)
    cloog_domain_print(foo,loop->domain) ;
    else
    fprintf(foo,"Null domain.\n") ;
    
    if (loop->inner != NULL)
    { fprintf(foo,"Inners loops are:\n") ;
      cloog_loop_print(foo,loop->inner) ;
    }
    else
    fprintf(foo,"No inners loops.\n") ;
    cloog_statement_print(foo,loop->statement) ;
    if (loop->next == NULL)
    fprintf(foo,"No next loops.\n") ;
    else
    fprintf(foo,"Next loop is:\n") ;
    fprintf(foo,"\n") ;

    loop = loop->next ;
  }
}


/******************************************************************************
 *                         Memory deallocation function                       *
 ******************************************************************************/


/* cloog_loop_free function:
 * This function frees the allocated memory for a CloogLoop structure (loop),
 * and frees its inner loops and its next loops.
 */
void cloog_loop_free(CloogLoop * loop)
{ CloogLoop * next ;
  
  while (loop != NULL)
  { cloog_loop_leak_down() ;
    
    next = loop->next ;
    cloog_domain_free(loop->domain) ;
    cloog_statement_free(loop->statement) ;
    if (loop->inner != NULL)
    cloog_loop_free(loop->inner) ;
    
    free(loop) ;
    loop = next ;
  }
}


/* cloog_loop_free_parts function:
 * This function frees the allocated memory for some parts of a CloogLoop
 * structure (loop), each other argument is a boolean having to be set to 1 if
 * we want to free the corresponding part, 0 otherwise. This function applies
 * the same freeing policy to its inner ans next loops recursively.
 * July 3rd 2003: first version.
 */
void cloog_loop_free_parts(loop, domain, statement, inner, next)
CloogLoop * loop ;
int domain, statement, inner, next ;
{ CloogLoop * follow ;

  while (loop != NULL)
  { cloog_loop_leak_down() ;
    follow = loop->next ;
    
    if (domain)
    cloog_domain_free(loop->domain) ;
    loop->domain = NULL ;
       
    if (statement)
    cloog_statement_free(loop->statement) ;
    loop->statement = NULL ;
     
    if ((inner) && (loop->inner != NULL))
    cloog_loop_free_parts(loop->inner,domain,statement,inner,next) ;
    loop->inner = NULL ;
    
    free(loop) ;
    if (next)
    loop = follow ;
    else
    loop = NULL ;
  }
}


/******************************************************************************
 *                              Reading functions                             *
 ******************************************************************************/


/* cloog_loop_read function:
 * This function reads loop data into a file (foo, posibly stdin) and
 * returns a pointer to a CloogLoop structure containing the read information.
 * This function can be used only for input file reading, when one loop is
 * associated with one statement.
 * - number is the number of the loop (i.e. the number of the corresponding
 *   statement).
 * - nb_parameters is the number of parameters.
 * September 9th 2002: first version.
 */
CloogLoop * cloog_loop_read(FILE * foo, int number, int nb_parameters)
{ int nb_iterators ;
  CloogLoop * loop ;

  cloog_loop_leak_up() ;
  
  /* Memory allocation and information reading for the first domain: */
  loop = (CloogLoop *)malloc(sizeof(CloogLoop)) ;
  if (loop == NULL) 
  { fprintf(stderr, "Memory Overflow.\n") ;
    exit(1) ;
  }
  /* domain. */
  loop->domain = cloog_domain_union_read(foo) ;
  if (loop->domain != NULL)
  nb_iterators = loop->domain->Dimension - nb_parameters ;
  else
  nb_iterators = 0 ;
  /* stride is initialized to 1. */
  loop->stride = 1 ;
  /* included statement. */
  loop->statement = cloog_statement_read(foo,nb_iterators,number) ;
  /* inner is NULL at beginning. */
  loop->inner = NULL ;
  /* next element. */
  loop->next = NULL ;

  return loop ;
}


/******************************************************************************
 *                            Processing functions                            *
 ******************************************************************************/


/* cloog_loop_alloc function:
 * This function allocates the memory space for a CloogLoop structure and
 * sets its fields with those given as input. Then it returns a pointer to the
 * allocated space.
 * October 27th 2001: first version. 
 */ 
CloogLoop * cloog_loop_alloc(domain, stride, statement, inner, next)
Polyhedron * domain ;
CloogStatement * statement ;
CloogLoop * inner, * next ;
{ CloogLoop * loop ;
  
  cloog_loop_leak_up() ;
  
  loop = (CloogLoop *)malloc(sizeof(CloogLoop)) ;
  if (loop == NULL)
  { fprintf(stderr, "Memory Overflow.\n") ;
    exit(1) ;
  }
  loop->domain = domain ;
  loop->stride = stride ;
  loop->statement = statement ;
  loop->inner = inner ;
  loop->next = next ;
  
  return(loop) ;
}


/* cloog_loop_add function:
 * This function adds a CloogLoop structure (loop) at a given place (now) of a
 * NULL terminated list of CloogLoop structures. The beginning of this list
 * is (start). This function updates (now) to (loop), and updates (start) if the
 * added element is the first one -that is when (start) is NULL-.
 * October 28th 2001: first version. 
 */ 
void cloog_loop_add(CloogLoop ** start, CloogLoop ** now, CloogLoop * loop)
{ if (*start == NULL)
  { *start = loop ;
    *now = *start ;
  }
  else
  { (*now)->next = loop ;
    *now = (*now)->next ;
  }
}


/* cloog_loop_copy function:
 * This function returns a copy of the CloogLoop structure given as input. In
 * fact, there is just new allocations for the CloogLoop structures, but their
 * contents are the same.
 * October 28th 2001: first version. 
 * July 3rd->11th 2003: memory leaks hunt and correction.
 */ 
CloogLoop * cloog_loop_copy(CloogLoop * source)
{ CloogLoop * loop ;
  CloogStatement * statement ;
  Polyhedron * domain ;

  loop = NULL ;
  if (source != NULL)
  { domain = cloog_domain_copy(source->domain) ;
    statement = cloog_statement_copy(source->statement) ;
    loop = cloog_loop_alloc(domain,source->stride,statement,NULL,NULL) ;
    loop->inner = cloog_loop_copy(source->inner) ;
    loop->next = cloog_loop_copy(source->next) ;
  }
  return(loop) ;
}


/* cloog_loop_add_separate function:
 * This function adds some CloogLoop structures at a given place (now) of a
 * NULL terminated list of CloogLoop structures. The beginning of this list
 * is (start). (loop) can be an union of polyhedra, this function separates the
 * union into a list of disjoint polyhedra then adds the list. This function
 * updates (now) to the end of the list and updates (start) if first added
 * element is the first of the principal list -that is when (start) is NULL-.
 * (loop) can be freed by this function, basically when its domain is actually
 * a union of polyhedra, but don't worry, all the useful data are now stored
 * inside the list (start).
 * October 28th 2001: first version. 
 * November 14th 2001: bug correction (this one was hard to find !).
 * July 3rd->11th 2003: memory leaks hunt and correction.
 */ 
void cloog_loop_add_separate(start, now, loop)
CloogLoop ** start, ** now, * loop ;
{ CloogLoop * sep, * inner ;
  Polyhedron * domain, * convex, * seen, * seen_before, * temp, * rest ;
  
  if (loop->domain->next == NULL)
  { loop->next = NULL ;
    cloog_loop_add(start,now,loop) ;
  }
  else
  { convex = cloog_domain_convex(loop->domain) ;
    temp   = cloog_domain_difference(convex,loop->domain) ;
    cloog_domain_free(loop->domain) ;
    loop->domain = NULL ;
    domain = cloog_domain_difference(convex,temp) ;
    cloog_domain_free(convex) ;
    cloog_domain_free(temp) ;

    /* We separate the first element of the rest of the union. */
    rest = domain->next ;
    domain->next = NULL ;

    /* This first element is the first of the list of disjoint polyhedra. */
    sep = cloog_loop_alloc(domain,1,loop->statement,loop->inner,NULL) ;
    cloog_loop_add(start,now,sep) ;
  
    /* If there are other elements, add a loop for each of them. */
    if (rest != NULL)
    { /* domain is used by the first element, and we will free 'seen', so... */
      seen = cloog_domain_copy(domain) ;

      while ((domain = rest) != NULL)
      { rest = domain->next ;
        domain->next = NULL ;
        temp = cloog_domain_difference(domain,seen) ;
    
        /* Each new loop will have its own life, for instance we can free its
         * inner loop. Then each one must have its own copy of 'inner'.
         */
        inner = cloog_loop_copy(loop->inner) ;
      
        sep = cloog_loop_alloc(temp,1,loop->statement,inner,NULL) ;
        /* temp can be an union too. If so: recursion. */
        if (temp->next != NULL)
        cloog_loop_add_separate(start,now,sep) ;
        else
        cloog_loop_add(start,now,sep) ;
    
        seen_before = seen ;
        if (rest != NULL)
        seen = cloog_domain_union(seen_before,domain) ;
        cloog_domain_free(seen_before) ;
      
        /* domain is allocated once, but freed in parts ! */
        cloog_domain_freed -- ;
        cloog_domain_free(domain) ;
      }
    }
    cloog_loop_free_parts(loop,0,0,0,0) ;  
  }
}


/* cloog_loop_restrict function:
 * This function returns the (loop) in the context of (context): it makes the
 * intersection between the (loop) domain and the (context), then it returns
 * a pointer to a new loop, with this intersection as domain.
 * - nb_par is the number of parameters.
 * October 27th 2001: first version. 
 */ 
CloogLoop * cloog_loop_restrict(loop, context, nb_par)
CloogLoop * loop ;
Polyhedron * context ;
int nb_par ;
{ int new_dimension ;
  Polyhedron * domain, * extended_context, * new_domain ;
  CloogLoop * new_loop ;
    
  domain = loop->domain ;
  if (domain->Dimension > context->Dimension)
  { new_dimension = domain->Dimension - nb_par ;
    extended_context = cloog_domain_extend(context,new_dimension,nb_par) ;
    new_domain = cloog_domain_intersection(extended_context,loop->domain) ;
    cloog_domain_free(extended_context) ;
  }
  else
  new_domain = cloog_domain_intersection(context,loop->domain) ;
  
  if (cloog_domain_empty(new_domain))
  return(NULL) ;
  else
  { new_loop = cloog_loop_alloc(new_domain,1,loop->statement,loop->inner,NULL) ;
    return(new_loop) ;
  }
}


/* cloog_loop_project function:
 * This function returns the projection of (loop) on the (level) first
 * dimensions (outer loops). It makes the projection of the (loop) domain,
 * then it returns a pointer to a new loop, with this projection as domain.
 * - nb_par is the number of parameters.
 * October 27th 2001: first version. 
 * July 3rd->11th 2003: memory leaks hunt and correction.
 */ 
CloogLoop * cloog_loop_project(CloogLoop * loop, int level, int nb_par)
{ Polyhedron * new_domain ;
  CloogLoop * new_loop, * copy ;

  copy = cloog_loop_alloc(loop->domain,loop->stride,loop->statement,
                          loop->inner,NULL) ;

  if ((loop->domain->Dimension-nb_par) == level)
  new_domain = cloog_domain_copy(loop->domain) ;  
  else
  new_domain = cloog_domain_project(loop->domain,level,nb_par) ;

  new_loop = cloog_loop_alloc(new_domain,1,NULL,copy,NULL) ;
    
  return(new_loop) ;
}


/* cloog_loop_concat function:
 * This function returns a pointer to the concatenation of the copies of the
 * CloogLoop lists given as input.
 * October 28th 2001: first version. 
 */
CloogLoop * cloog_loop_concat(CloogLoop * a, CloogLoop * b)
{ CloogLoop * loop, * temp ;

  loop = a  ;
  temp = loop ;
  if (loop != NULL)
  { while (temp->next != NULL)
    temp = temp->next ;
    temp->next = b ;
  }
  else
  loop = b ;
  
  return(loop) ;
} 


/* cloog_loop_separate function:
 * This function implements the Quillere algorithm for separation of multiple
 * loops: for a given set of polyhedra (loop), it computes a set of disjoint
 * polyhedra such that the unions of these sets are equal, and returns this set.
 * October 28th 2001: first version. 
 * November 14th 2001: elimination of some unused statements.
 * August 13th 2002: (debug) in the case of union of polyhedra for one loop,
 *                   redundant constraints are fired.
 * July 3rd->11th 2003: memory leaks hunt and correction.
 */ 
CloogLoop * cloog_loop_separate(CloogLoop * loop)
{ int first, quick_equal=0, quick_disjoint=0 ;
  CloogLoop * new_loop, * new_inner, * res, * now, * temp, * Q, 
            * inner, * old ;
  Polyhedron * UQ, * old_UQ, * domain ;
  
  /*if (loop == NULL)
  return NULL ;*/
  
  /* If there is only one loop in the list, we eliminate the constraints
   * non-shared by all parts of the polyhedra union, and we reject them at
   * the next level. This *seems* to bypass the limitation of the
   * cloog_domain_simplify (DomainSimplify) function for domain unions. As a
   * consequence, the 'if' conditional that separate unions will be in the
   * innermost loop.
   */
  if (loop->next == NULL)
  { temp = NULL ;
    res  = cloog_loop_alloc(loop->domain,1,loop->statement,loop->inner,NULL) ;
    cloog_loop_add_separate(&temp,&now,res) ;
    loop->domain = NULL ; /* Freed by cloog_loop_add_separate. */
    cloog_loop_free_parts(loop,0,0,0,0) ;
    return temp ;
  }
  
  UQ     = cloog_domain_copy(loop->domain) ;
  domain = cloog_domain_copy(loop->domain) ;
  res    = cloog_loop_alloc(domain,1,loop->statement,loop->inner,NULL) ;
  	  
  old = loop ;
  while((loop = loop->next) != NULL)
  { temp = NULL ;
    first = 1 ;
    
    /* For all Q, add Q-loop associated with the statements of Q alone,
     * and Q inter loop associated with the statements of Q and loop.
     */
    Q = res ;
    while (Q != NULL)
    { if (Q->statement == NULL)
      { /* Add (Q inter loop). */
        if((quick_disjoint=cloog_domain_quick_disjoint(Q->domain,loop->domain)))
	domain = NULL ;
	else
	{ if ((quick_equal = cloog_domain_quick_equal(Q->domain,loop->domain)))
	  domain = cloog_domain_copy(Q->domain) ;
          else
	  domain = cloog_domain_intersection(Q->domain,loop->domain) ;
	
          if (!cloog_domain_empty(domain))
          { new_inner = cloog_loop_concat(cloog_loop_copy(Q->inner),
                                          cloog_loop_copy(loop->inner)) ;
	    new_loop = cloog_loop_alloc(domain,1,NULL,new_inner,NULL) ;
            cloog_loop_add_separate(&temp,&now,new_loop) ;
          }
          else
          cloog_domain_free(domain) ;
        }
        
	/* Add (Q - loop). */
        if (quick_disjoint)
	domain = cloog_domain_copy(Q->domain) ;
	else
	{ if (quick_equal)
	  domain = NULL ;
	  else
	  domain = cloog_domain_difference(Q->domain,loop->domain) ;
	}
	
        if (!cloog_domain_empty(domain))
        { new_loop = cloog_loop_alloc(domain,1,NULL,Q->inner,NULL) ;
          cloog_loop_add_separate(&temp,&now,new_loop) ;
        }
        else
        { cloog_domain_free(domain) ;
	  /* If Q->inner is no more useful, we can free it. */
          inner = Q->inner ;
          Q->inner = NULL ;
          if (first) /* For the first Q, inner is also old->inner. */
          old->inner = NULL ;
          cloog_loop_free(inner) ;
        }
      }
      Q = Q->next ;
    }

    /* Add loop-UQ associated with the statements of loop alone.*/
    if (cloog_domain_quick_disjoint(loop->domain,UQ))
    domain = cloog_domain_copy(loop->domain) ;
    else
    { if (cloog_domain_quick_equal(loop->domain,UQ))
      domain = NULL ;
      else
      domain = cloog_domain_difference(loop->domain,UQ) ;
    }
    
    if (!cloog_domain_empty(domain))
    { new_loop = cloog_loop_alloc(domain,1,NULL,loop->inner,NULL) ;
      cloog_loop_add_separate(&temp,&now,new_loop) ;
    }
    else
    { cloog_domain_free(domain) ;
      /* If loop->inner is no more useful, we can free it. */
      cloog_loop_free(loop->inner) ;
    }
    
    loop->inner = NULL ;

    old_UQ = UQ ;
    if (loop->next != NULL)
    UQ = cloog_domain_union(UQ,loop->domain) ;

    cloog_domain_free(old_UQ) ;
    cloog_loop_free_parts(res,1,0,0,1) ;

    first = 0 ;
    res = temp ;
  }
  
  cloog_loop_free_parts(old,1,0,0,1) ;
  return(res) ;
}


/* cloog_loop_merge function:
 * This function is the 'soft' version of loop_separate if we are looking for
 * a code much simpler (and less efficicient). Here we merge loops if they have
 * common parts in the iteration space (if the intersection of their domains is
 * not empty), and let them isolated otherwise. This function returns the new
 * CloogLoop list.
 * October 29th 2001: first version. 
 * July 3rd->11th 2003: memory leaks hunt and correction.
 */ 
CloogLoop * cloog_loop_merge(CloogLoop * loop)
{ CloogLoop * res, * merge, * now, * Q, * P, * new_inner, * next, * old ;
  Polyhedron * new_domain, * temp ;

  if ((loop == NULL) || (loop->next == NULL))
  return loop ;
  
  /* First loop is added to the target list. */
  res = cloog_loop_alloc(loop->domain,1,loop->statement,loop->inner,NULL) ;
  old = loop ;
  /* Now the domain is in 'res' and it we be freed. */
  loop->domain = NULL ;
  
  /* And one by one, we see if we have to merge or to add the other loops. */
  while((loop = loop->next) != NULL)
  { merge = NULL ;
    P = cloog_loop_alloc(loop->domain,1,loop->statement,loop->inner,NULL) ;
    Q = res ;
    /* Now the domain is in 'P' and it we be freed. */
    loop->domain = NULL ;

    /* For each loop in the target list, if the intersection with the new loop
     * is empty, we can add the new loop directly, otherwise, we can merge then
     * add the fusion.
     */
    while (Q != NULL)
    { temp = cloog_domain_intersection(Q->domain,P->domain) ;
      next = Q->next ;
      if (cloog_domain_empty(temp))
      { cloog_domain_free(temp) ;
        cloog_loop_add_separate(&merge,&now,Q) ;
      }
      else 
      { cloog_domain_free(temp) ;
        new_inner = cloog_loop_concat(Q->inner,P->inner) ;
        temp = cloog_domain_union(P->domain,Q->domain) ;
        new_domain = cloog_domain_convex(temp) ;
        cloog_domain_free(temp) ;
        /* Q and P are no more used (but their content yes !).*/
        cloog_loop_free_parts(P,1,0,0,0) ;
        cloog_loop_free_parts(Q,1,0,0,0) ;
        P = cloog_loop_alloc(new_domain,1,NULL,new_inner,NULL) ;
      }
      Q = next ;
    }

    /* If there was merging, add it, otherwise add the loop lonely.
     * DEBUG : ici pas besoin de s'assurer que P->next est NULL (possible que
     * non si pas de fusion) car le dernier loop etudie a loop->next = NULL.
     */
    cloog_loop_add_separate(&merge,&now,P) ;
    res = merge ;
  }
  cloog_loop_free_parts(old,0,0,0,1) ;

  return (res);
}


/* October 29th 2001 */ 
int cloog_loop_number(CloogLoop * loop)
{ int nb_loops=0 ;

  while (loop != NULL)
  { nb_loops ++ ;
    loop = loop->next ;
  }
  return(nb_loops) ;
}


/* October 29th 2001 */ 
int cloog_loop_terminal_number(CloogLoop * loop)
{ int nb_terminal_loops=0 ;

  while (loop != NULL)
  { if (loop->statement != NULL)
    nb_terminal_loops ++ ;
    loop = loop->next ;
  }
  return(nb_terminal_loops) ;
}


/* cloog_loop_sort function:
 * Adaptation from LoopGen 0.4 by F. Quillere. This function sorts a list of
 * parameterized disjoint polyhedra, in order to not have lexicographic order
 * violation (see Quillere paper).
 */ 
CloogLoop * cloog_loop_sort(CloogLoop * loop, int level, int nb_par)
{ CloogLoop * res, * now, ** loop_array ;
  Polyhedron ** pols ;
  int i, nb_loops, * permut ;

  /* If there is only one loop, it's the end. */
  if ((nb_loops = cloog_loop_number(loop)) <= 1)
  return(loop) ;

  /* We have to allocate memory for some useful components:
   * - loop_array: the loop array,
   * - pols: the array of domains to sort,
   * - permut: will give us a possible sort (maybe not the only one).
   */
  loop_array = (CloogLoop **)malloc(nb_loops*sizeof(CloogLoop *)) ;
  pols = (Polyhedron **)malloc(nb_loops*sizeof(Polyhedron *)) ;
  permut = (int *)malloc(nb_loops*sizeof(int)) ;

  /* We fill up the loop and domain arrays. */
  for (i=0;i<nb_loops;i++,loop=loop->next)
  { loop_array[i] = loop ;
    pols[i] = loop_array[i]->domain ;
  }
  
  /* cloog_domain_sort will fill up permut. */
  cloog_domain_sort(pols,nb_loops,level,nb_par,permut) ;
  
  /* With permut and loop_array we build the sorted list. */
  res = NULL ;
  for (i=0;i<nb_loops;i++)
  { /* To avoid pointer looping... loop_add will rebuild the list. */
    loop_array[permut[i]-1]->next = NULL ;
    cloog_loop_add(&res,&now,loop_array[permut[i]-1]) ;
  }
  
  free(permut) ;
  free(pols) ;
  free(loop_array) ;

  return res;
}


/* cloog_loop_nest function:
 * This function changes the loop list in such a way that we have no more than
 * one dimension added by level. It returns an equivalent loop list with
 * this property.
 * October 29th 2001: first version.
 * July 3rd->11th 2003: memory leaks hunt and correction.
 */ 
CloogLoop * cloog_loop_nest(loop, context, level, nb_par)
CloogLoop * loop ;
Polyhedron * context ;
int level, nb_par ;
{ CloogLoop * p, * temp, * res, * now, * next ;
  Polyhedron * new_domain ;
  int l ;
  
  res = NULL ;
  /* Each domain is changed by its intersection with the context. */
  while (loop != NULL)
  { p = cloog_loop_restrict(loop,context,nb_par) ;
    next = loop->next ;
    cloog_loop_free_parts(loop,1,0,0,0) ;
    
    temp = cloog_loop_alloc(p->domain,1,p->statement,p->inner,NULL) ;
    
    /* If the intersection dimension is too big, we make projections smaller and
     * smaller, and each projection includes the preceding projection
     * (thus, in the target list, dimensions are added one by one).
     */
    if ((p->domain->Dimension-nb_par) > level)
    for (l=p->domain->Dimension-nb_par-1;l>=level;l--)
    { new_domain = cloog_domain_project(p->domain,l,nb_par) ;
      temp = cloog_loop_alloc(new_domain,1,NULL,temp,NULL) ;
    }
    /* p is no more useful (but its content yes !). */ 
    cloog_loop_free_parts(p,0,0,0,0) ;

    cloog_loop_add(&res,&now,temp) ;
    loop = next ;
  }
  return(res) ;
}


/* cloog_loop_stride function:
 * This function will find the stride of a loop for the iterator at the column
 * number 'level' in the constraint matrix. It will update the lower bound of
 * the iterator accordingly. Basically, the function will try to find in the
 * inner loops a common condition on this iterator for the inner loop iterators
 * to be integral. For instance, let us consider a loop with the iterator i,
 * the iteration domain -4<=i<=n, and its two inner loops with the iterator j.
 * The first inner loop has the constraint 3j=i, and the second one has the
 * constraint 6j=i. Then the common constraint on i for j to be integral is
 * i%3=0, the stride for i is 3. Lastly, we have to find the new lower bound
 * for i: the first value satisfying the common constraint: -3. At the end, the
 * iteration domain for i is -3<=i<=n and the stride for i is 3.
 * - loop is the loop including the iteration domain of the considered iterator,
 * - level is the column number of the iterator in the matrix of contraints.
 * June 29th 2003: first version (work in progress since June 26th 2003).
 * July 14th 2003: simpler version.
 */
void cloog_loop_stride(CloogLoop * loop, int level, int nb_par)
{ int first_search ;
  Value stride, ref_offset=0, offset=0, potential, lower=0 ;
  CloogLoop * inner ;

  /* Default stride. */
  stride = 1 ;
  first_search = 1 ;
  inner = loop->inner ;
    
  if (cloog_domain_integral_lowerbound(loop->domain,level,&lower))
  while (inner != NULL)
  { /* If the minimun stride has not been found yet, find the stride. */
    if ((first_search) || (stride != 1))
    { if (((potential=cloog_domain_stride(inner->domain,level,nb_par,
                                             &offset))!=1) && (!first_search))
      { /* Offsets must be the same. */
        if (offset == ref_offset)
	stride = cloog_domain_gcd(potential,stride) ;
        else
        stride = 1 ;
      }
      else
      { stride = potential ;
        ref_offset = offset ;
      }
	
      first_search = 0 ;
    }
      
    inner = inner->next ;
  }
    
  /* Update the values if necessary. */
  if (stride != 1)
  { /* Update the stride value. */
    loop->stride = stride ;
    /* Update the lower bound (too late to do something more intelligent !). */
    while (((lower-offset)%stride) != 0) 
    lower++ ;
    cloog_domain_lowerbound_update(loop->domain,level,lower) ;
  }
}


/* cloog_loop_generate function:
 * Adaptation from LoopGen 0.4 by F. Quillere. This function implements the
 * Quillere algorithm for polyhedron scanning. (see the Quillere paper).
 * October 26th 2001: first version.
 * July 31th 2002: (debug) there is no more parasite loops (REALLY hard bug !). 
 * July 3rd->11th 2003: memory leaks hunt and correction.
 */ 
CloogLoop * cloog_loop_generate(loop, context, level, nb_par, options)
CloogLoop * loop ;
Polyhedron * context ;
int level, nb_par ;
CloogOptions * options ;
{ CloogLoop * res, * now, * temp, * l, * new_loop, * inner, * now2, * end,
            * next, * next2, * into, * old /*, * points=NULL */;
  Polyhedron * domain ;

  res = NULL ;
  
  /* 1. Replace each polyhedron by its intersection with the context.
   * 2. Compute the projection of each polyhedron onto the outermost
   *    loop variable and the parameters.
   */
  while (loop != NULL)
  { temp = cloog_loop_restrict(loop,context,nb_par) ;
    if (temp != NULL)
    { old = temp ;
      temp = cloog_loop_project(temp,level,nb_par) ;
      cloog_loop_free_parts(old,0,0,0,0) ;
      cloog_loop_add(&res,&now,temp) ;
    }

    next = loop->next ;
    cloog_loop_free_parts(loop,1,0,0,0) ;
    loop = next ;
  }
  if (res == NULL)
  return NULL ;
  
  /* 3. Separate all projections into disjoint polyhedra. */
  res = ((options->f > level) || (options->f < 0)) ?
        cloog_loop_merge(res) : cloog_loop_separate(res) ;
  
  /* 3b. -correction- sort the loops to determine their textual order. */
  res = cloog_loop_sort(res,level,nb_par) ;
      
  /* 4. Recurse for each loop with the current domain as context. */
  temp = res ;
  res = NULL ;
  if ((level < options->l) || (options->l < 0))
  while(temp != NULL)
  { if (options->strides)
    cloog_loop_stride(temp,level,nb_par) ;
    
    inner = temp->inner ;
    domain = temp->domain ;
    into = NULL ;
    while (inner != NULL)
    { /* 4b. -ced- recurse for each sub-list of non terminal loops. */
      if (inner->domain->Dimension > (level + nb_par))
      { end = inner ;
        while ((end->next != NULL) &&
               (end->next->domain->Dimension > (level + nb_par)))
        end = end->next ;
        next = end->next ;
        end->next = NULL ;

        l = cloog_loop_generate(inner,domain,level+1,nb_par,options) ;
        if (l != NULL)
        cloog_loop_add(&into,&now,l) ;
        
        inner = next ;
      }
      else
      { cloog_loop_add(&into,&now,inner) ;
        inner = inner->next ;
      }
    }
    next = temp->next ;
    temp->next = NULL ;
    temp->inner = into ;
    cloog_loop_add(&res,&now2,temp) ;
    temp = next ;  
  }
  else
  while (temp != NULL)
  { next = temp->next ;
    l = cloog_loop_nest(temp->inner,temp->domain,level+1,nb_par) ;
    new_loop = cloog_loop_alloc(temp->domain,1,NULL,l,NULL) ;
    temp->inner = NULL ;
    temp->next = NULL ;
    cloog_loop_free_parts(temp,0,0,0,0) ;
    cloog_loop_add(&res,&now,new_loop) ;
    temp = next ;
  }
  
  /* 5. eliminate unused iterations of the current level for the new one. See
   *    the example called linearity-1-1 example with and without this part
   *    for an idea.
   */
  if (((level < options->l) || (options->l < 0)) &&
      ((options->f <= level) && !(options->f < 0)))
  { temp = res ;
    res = NULL ;

    while (temp != NULL)
    { l = NULL ;
      inner = temp->inner ;
      
      while (inner != NULL)
      { next = inner->next ;
        // This 'if' and its first part is the debug of july 31th 2002. 
        if (inner->statement != NULL)
        { end = cloog_loop_alloc(inner->domain,1,inner->statement,NULL,NULL) ;
          domain = cloog_domain_copy(temp->domain) ;
          new_loop = cloog_loop_alloc(domain,1,NULL,end,NULL) ;
        }
        else
        new_loop = cloog_loop_project(inner,level,nb_par) ;

        cloog_loop_free_parts(inner,0,0,0,0) ;
        cloog_loop_add(&l,&now2,new_loop) ;
        inner = next ;
      }
      
      temp->inner = NULL ;
      
      if (l != NULL)
      { l = cloog_loop_separate(l) ;
        l = cloog_loop_sort(l,level,nb_par) ;
        while (l != NULL)
        { l->stride = temp->stride ;
          cloog_loop_add(&res,&now,l) ;
          l = l->next ;
        }
      }
      next2 = temp->next ;
      cloog_loop_free_parts(temp,1,0,0,0) ;
      temp = next2 ;
    }
  }
  
  /* Pray for my new paper to be accepted somewhere since the following stuff
   * is really amazing :-) !
   */
  /* res = cloog_loop_unisolate(res,context,level,nb_par) ;*/
  
  return(res) ;
}


/* cloog_loop_simplify function:
 * This function implements the part 6. of the Quillere algorithm, it
 * recursively simplifies each loop in the context of the preceding loop domain.
 * It returns a pointer to the simplified loop list.
 * The cloog_domain_simplify (DomainSimplify) behaviour is really bad with
 * polyhedra union and some really awful sidesteppings were written, I plan
 * to solve that...
 * October 31th 2001: first version. 
 * November 23th 2001: a lot of changes. Before: i=M;j=i;k=j; now: i=M;j=M;k=M;
 *                   (as Quillere) not absolutely necessary, but so beautiful !
 * July 3rd->11th 2003: memory leaks hunt and correction.
 */ 
CloogLoop * cloog_loop_simplify(loop, context, level, nb_par)
CloogLoop * loop ;
Polyhedron * context ;
int level, nb_par ;
{ CloogLoop * simplified, * inner, * next ;
  CloogStatement * new_statement ;
  Polyhedron * domain, * rext_domain, * proj_domain, * miss, * exmiss, * old,
             * inter, * simp_inter, * simp, * proj_simp, * extended_context ;
	     
  if (loop == NULL)
  return(NULL) ;
  
  domain = loop->domain ;
  
  /* If the constraint system is never true, go to the next one. */
  while (cloog_domain_never_integral(domain,level,nb_par))
  { next = loop->next ;
    inner = loop->inner ;
    cloog_loop_free_parts(loop,0,1,0,0) ;
    cloog_loop_free(inner) ;
    
    if (next == NULL)
    return NULL ;
    else
    { loop = next ;
      domain = loop->domain ;
    }
  }

  next = cloog_loop_simplify(loop->next,context,level,nb_par) ;
  inner = cloog_loop_simplify(loop->inner,domain,level+1,nb_par) ;
  
  if ((inner == NULL) && (loop->statement == NULL))
  { loop->inner = NULL ; /* For loop integrity. */
    loop->next = NULL ;  /* For loop integrity. */
    cloog_loop_free_parts(loop,1,1,1,0) ;
    return(next) ;
  }
  
  new_statement = cloog_statement_copy(loop->statement) ;
  
  extended_context=cloog_domain_extend(context,domain->Dimension-nb_par,nb_par);
  /* To recover the old fast simple version : just
   * simp = cloog_domain_simplify(domain,extended_context) ;
   * and the last seven lines of the function.
   */ 
   
  /* Simplify the intersection between the domain and the context with the
   * domain 'without' the context dimensions. This will perform the
   * i=M;j=i; -> i=M;j=M; improvement.
   */
  inter = cloog_domain_intersection(domain,extended_context) ;
  proj_domain = cloog_domain_project(domain,context->Dimension-nb_par,nb_par) ;
  rext_domain=cloog_domain_extend(proj_domain,domain->Dimension-nb_par,nb_par) ;
  simp_inter = cloog_domain_simplify(inter,rext_domain) ;
  cloog_domain_free(inter) ;
  cloog_domain_free(rext_domain) ;

  /* Remove the redundant constraints. */
  simp = cloog_domain_simplify(simp_inter,extended_context) ;
  cloog_domain_free(simp_inter) ;
  cloog_domain_free(extended_context) ;
  
  /* Maybe too much constraints were simplified by the intersection
   * simplification. Check it and restore them if necessary.
   */
  proj_simp = cloog_domain_project(simp,context->Dimension-nb_par,nb_par) ;
  inter = cloog_domain_intersection(proj_simp,context) ;
  cloog_domain_free(proj_simp) ;
  miss = cloog_domain_simplify(proj_domain,inter) ;
  cloog_domain_free(proj_domain) ;
  cloog_domain_free(inter) ;
  
  if (!cloog_domain_universe(miss))
  { exmiss = cloog_domain_extend(miss,domain->Dimension-nb_par,nb_par) ;
    old = simp ;
    simp = cloog_domain_addconstraints(exmiss,old) ;
    cloog_domain_free(exmiss) ;
    cloog_domain_free(old) ;
  }
  cloog_domain_free(miss) ;
  
  simplified = cloog_loop_alloc(simp,loop->stride,new_statement,inner,next) ;

  loop->inner = NULL ; /* For loop integrity. */
  loop->next = NULL ;  /* For loop integrity. */
  cloog_loop_free_parts(loop,1,1,0,0) ;

  return(simplified) ; 
}


/* cloog_loop_fast_simplify function:
 * This function implements the part 6. of the Quillere algorithm, it
 * recursively simplifies each loop in the context of the preceding loop domain.
 * It returns a pointer to the simplified loop list.
 * The cloog_domain_simplify (DomainSimplify) behaviour is really bad with
 * polyhedra union and some really awful sidesteppings were written, I plan
 * to solve that...
 * October 31th 2001: first version. 
 * July 3rd->11th 2003: memory leaks hunt and correction.
 */ 
CloogLoop * cloog_loop_fast_simplify(loop, context, level, nb_par)
CloogLoop * loop ;
Polyhedron * context ;
int level, nb_par ;
{ CloogLoop * simplified, * inner, * next ;
  CloogStatement * new_statement ;
  Polyhedron * domain, * simp, * extended_context ;
	     
  if (loop == NULL)
  return(NULL) ;
  
  domain = loop->domain ;
  
  /* If the constraint system is never true, go to the next one. */
  while (cloog_domain_never_integral(domain,level,nb_par))
  { next = loop->next ;
    inner = loop->inner ;
    cloog_loop_free_parts(loop,0,1,0,0) ;
    cloog_loop_free(inner) ;
    
    if (next == NULL)
    return NULL ;
    else
    { loop = next ;
      domain = loop->domain ;
    }
  }

  next = cloog_loop_fast_simplify(loop->next,context,level,nb_par) ;
  inner = cloog_loop_fast_simplify(loop->inner,domain,level+1,nb_par) ;
  
  if ((inner == NULL) && (loop->statement == NULL))
  { loop->inner = NULL ; /* For loop integrity. */
    loop->next = NULL ;  /* For loop integrity. */
    cloog_loop_free_parts(loop,1,1,1,0) ;
    return(next) ;
  }
  
  new_statement = cloog_statement_copy(loop->statement) ;
  
  extended_context=cloog_domain_extend(context,domain->Dimension-nb_par,nb_par);
  simp = cloog_domain_simplify(domain,extended_context) ;

  simplified = cloog_loop_alloc(simp,loop->stride,new_statement,inner,next) ;

  loop->inner = NULL ; /* For loop integrity. */
  loop->next = NULL ;  /* For loop integrity. */
  cloog_loop_free_parts(loop,1,1,0,0) ;

  return(simplified) ; 
}


/* cloog_loop_scatter function:
 * This function add the scattering (scheduling) informations in a loop.
 * November 4th 2001: first version. 
 */
void cloog_loop_scatter(CloogLoop * loop, Polyhedron * scatt)
{ int scatt_dim ;
  Polyhedron * domain, * ext, * newdom, * newpart, * temp ;
  
  scatt_dim = scatt->Dimension - loop->domain->Dimension ;
  domain = loop->domain ;
  newdom = NULL ;
  
  /* For each polyhedron of domain (it can be an union of polyhedra). */
  while (domain != NULL)
  { /* Extend the domain to by adding the scattering dimensions as the new
     * first domain dimensions.
     */
    ext = cloog_domain_extend(domain,scatt_dim,domain->Dimension) ;
    /* Then add the scattering constraints. */
    newpart = cloog_domain_addconstraints(scatt,ext) ;
    cloog_domain_free(ext) ;

    if (newdom != NULL)
    { temp = newdom ;
      newdom = cloog_domain_union(newdom,newpart) ;
      cloog_domain_free(temp) ;
      cloog_domain_free(newpart) ;
    }
    else
    newdom = newpart ;

    temp = domain ;
    domain = domain->next ;
    temp->next = NULL ; /* We don't want to free the rest of the list. */
    cloog_domain_free(temp) ;
  }
  
  loop->domain = newdom ;
}
